package com.flyfish.framework.r2dbc.config.callback;

import com.flyfish.framework.domain.base.Domain;
import com.flyfish.framework.domain.base.Qo;
import com.flyfish.framework.query.Queries;
import com.flyfish.framework.r2dbc.metadata.R2dbcMetadataManager;
import com.flyfish.framework.r2dbc.metadata.R2dbcTableMetadata;
import com.flyfish.framework.r2dbc.metadata.reference.FieldSetter;
import com.flyfish.framework.r2dbc.metadata.reference.R2dbcAssociation;
import com.flyfish.framework.r2dbc.metadata.reference.R2dbcCollection;
import com.flyfish.framework.r2dbc.metadata.reference.R2dbcRelation;
import com.flyfish.framework.repository.DefaultReactiveRepository;
import lombok.RequiredArgsConstructor;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.ObjectUtils;
import org.reactivestreams.Publisher;
import org.springframework.beans.factory.ObjectProvider;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.mapping.PersistentEntity;
import org.springframework.data.mapping.PersistentProperty;
import org.springframework.data.mapping.PersistentPropertyAccessor;
import org.springframework.data.r2dbc.core.R2dbcEntityOperations;
import org.springframework.data.r2dbc.core.StatementMapper;
import org.springframework.data.r2dbc.mapping.event.AfterConvertCallback;
import org.springframework.data.relational.core.mapping.RelationalPersistentProperty;
import org.springframework.data.relational.core.query.Criteria;
import org.springframework.data.relational.core.sql.SqlIdentifier;
import org.springframework.data.util.CastUtils;
import org.springframework.data.util.Lazy;
import org.springframework.lang.NonNull;
import org.springframework.r2dbc.core.DatabaseClient;
import org.springframework.r2dbc.core.PreparedOperation;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.function.Function;
import java.util.stream.Collectors;

@RequiredArgsConstructor
public class ReferenceR2dbcCallback implements AfterConvertCallback<Domain> {

    private final R2dbcMetadataManager r2dbcMetadataManager;
    private Lazy<R2dbcEntityOperations> operations;
    private Lazy<Map<Class<?>, DefaultReactiveRepository<?>>> repositories;

    @Override
    @NonNull
    public Publisher<Domain> onAfterConvert(@NonNull Domain entity, @NonNull SqlIdentifier table) {
        return Mono.deferContextual(ctx -> {
            Optional<Qo<? extends Domain>> query = ctx.getOrEmpty(Qo.KEY);
            query.ifPresent(qo -> entity.setCurrentUser(qo.getUser()));
            // 判断是否需要获取关联信息
            boolean fetchRefs = query.map(Qo::isFetchRef).orElse(false);
            if (fetchRefs) {
                return doFetch(entity);
            }
            return Mono.just(entity);
        });
    }

    /**
     * 完成填充
     *
     * @param entity 实体
     * @return 结果
     */
    private Mono<Domain> doFetch(Domain entity) {
        Class<? extends Domain> entityClass = entity.getClass();
        // 遍历fields，找到要注入的数据
        R2dbcTableMetadata metadata = r2dbcMetadataManager.getMetadata(entityClass);
        // 获取本实体的持久化对象
        PersistentEntity<?, ?> persistentEntity = metadata.getEntity();
        // 获取属性设置器
        PersistentPropertyAccessor<Domain> accessor = persistentEntity.getPropertyAccessor(entity);

        // 尝试填充各种关联方式
        List<Mono<Domain>> signals = new ArrayList<>();
        signals.addAll(fetchAssociation(metadata.getAssociations(), accessor));
        signals.addAll(fetchCollections(metadata.getCollections(), accessor));
        signals.addAll(fetchRelations(metadata.getRelations(), accessor));

        // 尝试填充
        if (CollectionUtils.isNotEmpty(signals)) {
            return Mono.zip(signals, objs -> entity);
        }
        return Mono.just(entity);
    }

    private DefaultReactiveRepository<Domain> getRepository(Class<?> entityClass) {
        return CastUtils.cast(repositories.get().get(entityClass));
    }

    @Autowired
    public void setRepositories(ObjectProvider<DefaultReactiveRepository<?>> repositories) {
        this.repositories = Lazy.of(() -> repositories.stream()
                .collect(Collectors.toMap(repo -> repo.getEntityInformation().getJavaType(), Function.identity())));
    }

    @Autowired
    public void setR2dbcEntityOperations(ObjectProvider<R2dbcEntityOperations> operations) {
        this.operations = Lazy.of(operations::getIfAvailable);
    }

    /**
     * 拉取一对一关联
     *
     * @param associations 一对一关联集合
     * @param accessor     本对象的属性访问器
     * @return 结果
     */
    private List<Mono<Domain>> fetchAssociation(List<R2dbcAssociation> associations, PersistentPropertyAccessor<Domain> accessor) {
        List<Mono<Domain>> signals = new ArrayList<>();
        if (CollectionUtils.isNotEmpty(associations)) {
            // 分区，将判断提在外面，提升性能
            Map<Boolean, List<R2dbcAssociation>> partitions = associations.stream()
                    .collect(Collectors.partitioningBy(R2dbcAssociation::isInner));
            // 处理内部查询
            if (CollectionUtils.isNotEmpty(partitions.get(true))) {
                partitions.get(true).forEach(association -> {
                    // 获取对方的仓库
                    DefaultReactiveRepository<Domain> repository = getRepository(association.getEntity().getType());
                    // 内部查询，以对象内的值作为查询条件
                    Object value = accessor.getProperty(association.getField());
                    if (ObjectUtils.isEmpty(value)) return;
                    // 准备设置
                    FieldSetter setter = association.getSetter();
                    // 查询
                    Mono<Domain> signal = repository.findById(String.valueOf(value))
                            .map(result -> setter.setValue(accessor.getBean(), result))
                            .defaultIfEmpty(accessor.getBean());
                    signals.add(signal);
                });
            }
            // 处理外部查询
            if (CollectionUtils.isNotEmpty(partitions.get(false))) {
                // 外部查询，以当前实体的id为条件，去外部表的字段上查询
                partitions.get(false).forEach(association -> {
                    // 获取条件
                    String id = accessor.getBean().getId();
                    // 准备查询
                    DefaultReactiveRepository<Domain> repository = getRepository(association.getEntity().getType());
                    // 准备设置
                    FieldSetter setter = association.getSetter();
                    // 直接通过外部字段查询
                    Mono<Domain> signal = repository.findOne(Queries.where(getColumnName(association.getField())).eq(id).wrap())
                            .map(result -> setter.setValue(accessor.getBean(), result))
                            .defaultIfEmpty(accessor.getBean());
                    signals.add(signal);
                });
            }
        }
        return signals;
    }

    /**
     * 拉取一对多关联实体
     *
     * @param collections 一对多关联
     * @param accessor    本对象的属性访问器
     * @return 结果
     */
    private List<Mono<Domain>> fetchCollections(List<R2dbcCollection> collections,
                                                PersistentPropertyAccessor<Domain> accessor) {
        List<Mono<Domain>> signals = new ArrayList<>();
        if (CollectionUtils.isNotEmpty(collections)) {
            // 外部查询，以当前实体的id为条件，去外部表的字段上查询
            collections.forEach(collection -> {
                // 获取条件
                String id = accessor.getBean().getId();
                // 准备查询
                DefaultReactiveRepository<Domain> repository = getRepository(collection.getEntity().getType());
                // 准备设置
                FieldSetter setter = collection.getSetter();
                // 直接通过外部字段查询
                Mono<Domain> signal = repository.findAll(Queries.where(getColumnName(collection.getField())).eq(id).wrap())
                        .collectList()
                        .map(result -> setter.setValue(accessor.getBean(), result))
                        .defaultIfEmpty(accessor.getBean());
                signals.add(signal);
            });
        }
        return signals;
    }

    /**
     * 拉取多对多关联数据
     *
     * @param relations 多对多关联信息
     * @param accessor  本对象的属性访问器
     * @return 结果
     */
    private List<Mono<Domain>> fetchRelations(List<R2dbcRelation> relations, PersistentPropertyAccessor<Domain> accessor) {
        List<Mono<Domain>> signals = new ArrayList<>();
        if (CollectionUtils.isNotEmpty(relations)) {
            // 生成DatabaseClient，执行sql语句查询
            relations.forEach(relation -> {
                // 准备目标对象的repo
                DefaultReactiveRepository<Domain> repository = getRepository(relation.getEntity().getType());
                // 准备查询参数
                String id = accessor.getBean().getId();
                // 查询关联id集合
                Flux<String> ids = fetchRelationIds(relation, id);
                // 准备设置
                FieldSetter setter = relation.getSetter();
                // 查询最终结果
                Mono<Domain> signal = repository.findAllById(ids)
                        .collectList()
                        .map(result -> setter.setValue(accessor.getBean(), result))
                        .defaultIfEmpty(accessor.getBean());
                signals.add(signal);
            });
        }
        return signals;
    }

    /**
     * 查询当前实体关联的ids
     *
     * @param relation 关联信息
     * @return 结果
     */
    private Flux<String> fetchRelationIds(R2dbcRelation relation, String id) {
        // 准备查询器
        R2dbcEntityOperations entityOperations = operations.get();
        StatementMapper statementMapper = entityOperations.getDataAccessStrategy().getStatementMapper();
        DatabaseClient databaseClient = entityOperations.getDatabaseClient();
        // 构建查询
        StatementMapper.SelectSpec select = statementMapper.createSelect(relation.getTableName())
                .withProjection(relation.getForeignField()).withCriteria(Criteria.where(relation.getField()).is(id));
        PreparedOperation<?> operation = statementMapper.getMappedObject(select);
        // 执行查询
        return databaseClient.sql(operation)
                .map(row -> row.get(0, String.class))
                .all();
    }

    /**
     * 通过持久化属性获取真实的列名
     *
     * @param property 属性
     * @return 结果
     */
    private String getColumnName(PersistentProperty<?> property) {
        if (property instanceof RelationalPersistentProperty) {
            return ((RelationalPersistentProperty) property).getColumnName().getReference();
        }
        return property.getName();
    }
}
